import unittest
import larry as lry

ENVIRONMENT_PROD = 'production'
ENVIRONMENT_SANDBOX = 'sandbox'
SANDBOX_HIT = '39HYCOOPKNK26VOMWWPV050D1O9MD5'
SANDBOX_HIT_TYPE = '3W679PTMVMW4B1YPP05F1CL2SYKBXP'
SANDBOX_ASSIGNMENT = '3TEM0PF1Q5W8Q0F8XU7ZRSPG1ARD0O'
PROD_HIT = '30Y6N4AHYOVT3B1E15NSX07Z8YNRDS'
PROD_HIT_TYPE = '32CVJ4DS80UD0FXOVYK5MQJIWDSKV8'
PROD_ASSIGNMENT = '3N4BPTXIO8RWKSXYNI9LV8K4SNYUK5'
SIMPLE_QUESTION = '<script src="https://assets.crowd.aws/crowd-html-elements.js"></script><crowd-form><p>What is the date today?</p><input name="date"></crowd-form>'
SIMPLE_TEMPLATE = '<script src="https://assets.crowd.aws/crowd-html-elements.js"></script><crowd-form><p>What day of the week was {{ date }}?</p><input name="date"></crowd-form>'
SIMPLE_TEMPLATE_URI = 's3://larry-testing/test-objects/mturk/simple_template.html'
BASIC_ANNOTATION_DICT = {'path': 'detail'}
BASIC_ANNOTATION_STRING = 'For easier data science use Larry'

EXTERNAL_URL = 'https://www.google.com'


class MTurkTests(unittest.TestCase):

    def test_use_production(self):
        lry.mturk.use_production()
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_PROD)
        self.assertTrue(lry.mturk.production())
        self.assertFalse(lry.mturk.sandbox())

    def test_use_sandbox(self):
        lry.mturk.use_sandbox()
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_SANDBOX)
        self.assertTrue(lry.mturk.sandbox())
        self.assertFalse(lry.mturk.production())

    def test_set_environment_prod(self):
        lry.mturk.set_environment('prod')
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_PROD)
        self.assertTrue(lry.mturk.production())
        self.assertFalse(lry.mturk.sandbox())

    def test_set_environment_sandbox(self):
        lry.mturk.set_environment('sandbox')
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_SANDBOX)
        self.assertTrue(lry.mturk.sandbox())
        self.assertFalse(lry.mturk.production())

    def test_set_environment_prod_hit(self):
        lry.mturk.set_environment(hit_id=PROD_HIT)
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_PROD)
        self.assertTrue(lry.mturk.production())
        self.assertFalse(lry.mturk.sandbox())

    def test_set_environment_sandbox_hit(self):
        lry.mturk.set_environment(hit_id=SANDBOX_HIT)
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_SANDBOX)
        self.assertTrue(lry.mturk.sandbox())
        self.assertFalse(lry.mturk.production())

    def test_set_environment_prod_assignment(self):
        lry.mturk.set_environment(assignment_id=PROD_ASSIGNMENT)
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_PROD)
        self.assertTrue(lry.mturk.production())
        self.assertFalse(lry.mturk.sandbox())

    def test_set_environment_sandbox_assignment(self):
        lry.mturk.set_environment(assignment_id=SANDBOX_ASSIGNMENT)
        self.assertEqual(lry.mturk.environment(), ENVIRONMENT_SANDBOX)
        self.assertTrue(lry.mturk.sandbox())
        self.assertFalse(lry.mturk.production())

    def test_create_hit(self):
        lry.mturk.use_sandbox()
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward_cents=10, lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   html_question=SIMPLE_QUESTION, annotation=BASIC_ANNOTATION_DICT)
        self.assertFalse(hit.production)
        hit = lry.mturk.get_hit(hit.hit_id)
        self.assertEqual(hit.annotation, BASIC_ANNOTATION_DICT)
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward_cents=10, lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   html_question=SIMPLE_QUESTION, annotation=BASIC_ANNOTATION_STRING)
        self.assertFalse(hit.production)
        hit = lry.mturk.get_hit(hit.hit_id)
        self.assertEqual(hit.annotation, BASIC_ANNOTATION_STRING)
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward='0.10', lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   question=lry.mturk.render_html_question(SIMPLE_QUESTION))
        self.assertFalse(hit.production)
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward='0.10', lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   question=lry.mturk.render_external_question(EXTERNAL_URL))
        self.assertFalse(hit.production)
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward='0.10', lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   external_question=EXTERNAL_URL)
        self.assertFalse(hit.production)
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward='0.10', lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   question_template=SIMPLE_TEMPLATE, template_context={'date': '2/13/2020'})
        self.assertFalse(hit.production)
        hit = lry.mturk.create_hit("Simple task", "Answer a simple question", reward='0.10', lifetime=60,
                                   assignment_duration=60, max_assignments=1, auto_approval_delay=600,
                                   question_template_uri=SIMPLE_TEMPLATE_URI, template_context={'date': '2/13/2020'})
        self.assertFalse(hit.production)

    def test_create_by_hit_type(self):
        lry.mturk.use_sandbox()
        hit_type_id = lry.mturk.create_hit_type(title="Simple task", description="Answer a simple question",
                                                reward="0.10", assignment_duration=60)
        hit_type_id = lry.mturk.create_hit_type(title="Simple task", description="Answer a simple question",
                                                reward_cents=10, assignment_duration=60, auto_approval_delay=60,
                                                keywords='foo,bar')
        hit = lry.mturk.create_hit(hit_type_id=hit_type_id, lifetime=60, max_assignments=1,
                                   html_question=SIMPLE_QUESTION)
        self.assertFalse(hit.production)


if __name__ == '__main__':
    unittest.main()



import json
from larry import utils
import boto3


client = None
# A local instance of the boto3 session to use
__session = boto3.session.Session()


def set_session(aws_access_key_id=None,
                aws_secret_access_key=None,
                aws__session_token=None,
                region_name=None,
                profile_name=None,
                boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws__session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session, client
    __session = boto_session if boto_session is not None else boto3.session.Session(**utils.copy_non_null_keys(locals()))
    client = __session.client('sqs')


def send_message(destination, message, sqs_client=None):
    sqs_client = sqs_client if sqs_client else client
    if type(message) == dict:
        message = json.dumps(message)
    return sqs_client.send_message(QueueUrl=destination, MessageBody=message)



from collections import UserDict, Mapping
from larry.utils import image


class Box(UserDict):
    def __init__(self, *data):
        UserDict.__init__(self)
        for dat in data:
            self.update(dat)

    @classmethod
    def from_obj(cls, obj):
        result = cls(obj)
        if 'top' in obj and 'left' in obj and 'width' in obj and 'height' in obj:
            result._calc_coordinates()
        elif 'coordinates' in obj and len(obj['coordinates']) == 4:
            result._calc_position()
        return result

    @classmethod
    def from_position(cls, position, **kwargs):
        obj = cls(position, kwargs)
        obj._calc_coordinates()
        return obj

    @classmethod
    def from_percentage(cls, coordinates, origin='topleft', size=None, width=None, height=None, **kwargs):
        if size is None and (height is None or width is None):
            raise Exception("Size or height and width are required")
        (width, height) = size if size else (width, height)

        if origin == 'bottomleft':
            coordinates = [coordinates[0] * width, (1 - coordinates[1]) * height,
                           coordinates[2] * width, (1 - coordinates[3]) * height]
        else:
            coordinates = [coordinates[0] * width, coordinates[1] * height,
                           coordinates[2] * width, coordinates[3] * height]
        obj = cls({'coordinates': coordinates}, kwargs)
        obj._calc_position()
        return obj

    @classmethod
    def from_coords(cls, coordinates, origin='topleft', size=None, width=None, height=None, **kwargs):
        if origin == 'bottomleft':
            if size is None and height is None:
                raise Exception("Using an origin of 'bottomleft' requires a size or height be provided")
            (width, height) = size if size else (None, height)
            coordinates = [coordinates[0], height - coordinates[3], coordinates[2], height - coordinates[1]]
        obj = cls({'coordinates': coordinates}, kwargs)
        obj._calc_position()
        return obj

    @staticmethod
    def is_box(obj):
        return isinstance(obj, Mapping) and (('top' in obj and 'left' in obj and 'width' in obj and 'height' in obj) or
                                             ('coordinates' in obj and len(obj['coordinates']) == 4))

    def scaled(self, ratio):
        box = self.copy()
        for key, value in box.items():
            if key == 'coordinates':
                box[key] = [v * ratio for v in value]
            elif key in ['top', 'left', 'width', 'height']:
                box[key] = value * ratio
        return box

    def offset(self, x, y):
        box = self.copy()
        for k, v in box.items():
            if k == 'coordinates':
                box[k] = [v[0] + x, v[1] + y, v[2] + x, v[3] + y]
            elif k == 'left':
                box[k] = v + x
            elif k == 'top':
                box[k] = v + y
        return box

    def intersecting_boxes(self, boxes, min_overlap=0):
        intersecting = []
        for box in boxes:
            if image.box_area(image.box_intersection(self, box)) > image.box_area(box)*min_overlap:
                intersecting.append(box.offset(-self['left'], -self['top']))
        return intersecting

    def area(self):
        return self['width'] * self['height']

    def _calc_coordinates(self):
        self['coordinates'] = [self['left'], self['top'], self['left'] + self['width'] - 1,
                               self['top'] + self['height'] - 1]

    def _calc_position(self):
        self['left'] = self['coordinates'][0]
        self['top'] = self['coordinates'][1]
        self['width'] = self['coordinates'][2] - self['coordinates'][0] + 1
        self['height'] = self['coordinates'][3] - self['coordinates'][1] + 1

    def __missing__(self, key):
        if key == 'coordinates':
            self._calc_coordinates()
            return self[key]
        elif key in ['top', 'left', 'width', 'height']:
            self._calc_position()
            return self[key]
        else:
            raise KeyError(key)



import boto3
from larry import s3
from larry import mturk
from larry import sqs
from larry import sts
from larry import sagemaker
from larry import lmbda
from larry import iam

__version__ = "0.1.21"


def _propagate_session():
    s3.set_session(boto_session=__session)
    mturk.set_session(boto_session=__session)
    sqs.set_session(boto_session=__session)
    sts.set_session(boto_session=__session)
    sagemaker.set_session(boto_session=__session)
    lmbda.set_session(boto_session=__session)
    iam.set_session(boto_session=__session)


# A local instance of the boto3 session to use
__session = boto3.session.Session()
_propagate_session()


def session():
    global __session
    if __session is None:
        __session = boto3.session.Session()
    return __session


def set_session(aws_access_key_id=None,
                aws_secret_access_key=None,
                aws_session_token=None,
                region_name=None,
                profile_name=None,
                boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws_session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session
    __session = boto_session if boto_session is not None else boto3.session.Session(**utils.copy_non_null_keys(locals()))
    _propagate_session()




import json
from datetime import datetime, timedelta
from larry.mturk.HIT import HIT
from larry.mturk.Assignment import Assignment
from larry.types import Box
import re


DATE_FORMAT = '%Y-%m-%d %H:%M:%S%z'
TIME_FORMAT = '%H:%M:%S'


class JSONEncoder(json.JSONEncoder):

    def default(self, obj):
        if isinstance(obj, datetime):
            return date_to_string(obj)
        elif isinstance(obj, timedelta):
            return round(obj.total_seconds(), 3)
        elif isinstance(obj, HIT):
            hit = {i: obj[i] for i in obj if i != 'Question'}
            hit['__HIT__'] = True
            return hit
        elif isinstance(obj, Assignment):
            assignment = {i: obj[i] for i in obj}
            assignment['__Assignment__'] = True
            return assignment
        elif isinstance(obj, Box):
            box = {i: obj[i] for i in obj}
            box['__Box__'] = True
            return box
        return json.JSONEncoder.default(self, obj)


def JSONDecoder(dct):
    if '__HIT__' in dct:
        return HIT(dct)
    if '__Assignment__' in dct:
        return Assignment(dct)
    if '__Box__' in dct:
        return Box(dct)
    return dct


def json_loads(value):
    return json.loads(value, object_hook=JSONDecoder)


def json_dumps(value, separators=None):
    if separators:
        return json.dumps(value, cls=JSONEncoder, separators=separators)
    else:
        return json.dumps(value, cls=JSONEncoder)


def date_to_string(obj):
    return obj.strftime(DATE_FORMAT)


def parse_date(obj):
    return datetime.strptime(obj, DATE_FORMAT)


def copy_non_null_keys(param_list):
    result = {}
    for key, val in param_list.items():
        if val is not None:
            result[key] = val
    return result


def map_parameters(parameters, key_map):
    result = {}
    for k, i in key_map.items():
        if parameters.get(k) is not None:
            result[i] = parameters[k]
    return result


def create_s3_key(path, extension):
    """Given a path and a file type, strips out all non A-Z0-9 characters and
    appends the file type to the end: making paths and urls s3 friendly"""
    return re.sub(r'\W+', '', path) + '.' + extension


def show_progress(index, every, lst):
    if index == len(lst)-1:
        print('All {} completed at {}'.format(index+1, datetime.now()))
    elif (index+1) % every == 0:
        print('Completed {} at {}'.format(index+1, datetime.now()))



import unittest
import larry as ld


class MyTestCase(unittest.TestCase):
    def test_something(self):
        ld.sagemaker.labeling.describe_job('foo')
        self.assertEqual(True, False)


if __name__ == '__main__':
    unittest.main()



from larry import utils
from larry import mturk
import collections


class HIT(collections.UserDict):

    def __init__(self, data, mturk_client=None, production=None):
        self.__client = mturk_client
        collections.UserDict.__init__(self)
        if isinstance(data, str):
            hit, prod = mturk._get_hit(data, mturk_client)
            self.update(hit)
            self['Production'] = prod
        else:
            self.update(data)
            self._parse_datetime_values()
            if production is not None:
                self['Production'] = production

    def _parse_datetime_values(self):
        for key in ['CreationTime', 'Expiration']:
            try:
                if key in self and isinstance(self[key], str):
                    self[key] = utils.parse_date(self[key])
            except ValueError:
                pass

    def __repr__(self):
        return "{}('{}')".format(type(self).__name__, self.hit_id)

    def __str__(self):
        return """{}:
    Production: {}
    Title: {}
    HIT Type: {}
    HIT Group: {}
    Status: {}
    Review Status: {}
    Reward: {}
    Pending: {}
    Available: {}
    Completed: {}""".format(self.hit_id, self.production, self.title, self.hit_type_id, self.hit_group_id,
                            self.status, self.review_status, self.reward, self.pending, self.available, self.completed)

    def refresh(self, get_assignments=False):
        self.update(mturk._get_hit(self.hit_id, self.__client)[0])
        if 'Assignments' in self or get_assignments:
            self.retrieve_assignments()

    def retrieve_assignments(self):
        self['Assignments'] = list(mturk.list_assignments_for_hit(self.hit_id))
        return self['Assignments']

    def retrieve_annotation(self, s3_resource=None):
        self['Annotation'] = mturk.parse_requester_annotation(self.get('RequesterAnnotation'),
                                                              s3_resource=s3_resource)

    def __missing__(self, key):
        if key == 'Assignments':
            self.retrieve_assignments()
            return self['Assignments']
        elif key == 'Annotation':
            self.retrieve_annotation()
            return self['Annotation']
        else:
            raise KeyError(key)

    @property
    def assignments(self):
        return self['Assignments']

    @property
    def answers(self):
        for assignment in self['Assignments']:
            answer = assignment['Answer'].copy()
            answer['WorkerId'] = assignment['WorkerId']
            yield answer

    @property
    def hit_id(self):
        return self['HITId']

    @property
    def reward(self):
        return float(self['Reward'])

    @property
    def reward_cents(self):
        return float(self['Reward']) * 100

    @property
    def hit_type_id(self):
        return self['HITTypeId']

    @property
    def hit_group_id(self):
        return self['HITGroupId']

    @property
    def hit_layout_id(self):
        return self.get('HITLayoutId', None)

    @property
    def creation_time(self):
        return self['CreationTime']

    @property
    def title(self):
        return self['Title']

    @property
    def description(self):
        return self['Description']

    @property
    def question(self):
        return self.get('Question', None)

    @property
    def keywords(self):
        return self['Keywords']

    @property
    def status(self):
        return self['HITStatus'] #: 'Assignable' | 'Unassignable' | 'Reviewable' | 'Reviewing' | 'Disposed',

    @property
    def max_assignments(self):
        return self.get('MaxAssignments', None)

    @property
    def auto_approval_delay(self):
        return self.get('AutoApprovalDelayInSeconds', None)

    @property
    def expiration(self):
        return self.get('Expiration', None)

    @property
    def duration(self):
        return self.get('AssignmentDurationInSeconds', None)

    @property
    def annotation(self):
        return self.get('Annotation', None)

    @property
    def qualification_requirements(self):
        return self.get('QualificationRequirements', None)

    @property
    def review_status(self):
        return self.get('HITReviewStatus', None)

    @property
    def pending(self):
        return self.get('NumberOfAssignmentsPending', None)

    @property
    def available(self):
        return self.get('NumberOfAssignmentsAvailable', None)

    @property
    def completed(self):
        return self.get('NumberOfAssignmentsCompleted', None)

    @property
    def production(self):
        return self.get('Production', None)

    @property
    def preview(self):
        return mturk.preview_url(self.hit_type_id, self.production)

    @property
    def completed_assignment_count(self):
        if 'Assignments' in self:
            return len(self['Assignments'])
        else:
            return 0

    def expire(self):
        mturk.expire_hit(self.hit_id, self.__client)



import unittest
import larry as lry
from larry.types import Box
from larry.utils import json_dumps
import datetime
from botocore.exceptions import ClientError

# S3 testing objects
SIMPLE_DICT = {
    'a': {'key': 'value'},
    'b': ['a', 'b', 'c'],
    '1': 15,
    'box': Box.from_coords([3,4,5,6]),
    'date': datetime.datetime.now()
}
SIMPLE_LIST_OF_DICTS = [
    {'a': 124, 'b': 'value'},
    {'a': 939, 'b': 'foo'},
    {'a': 389, 'b': 'bar', '3': 'new'}
]
SIMPLE_LIST = ['a', 'b', 'c', '3', 'd', '2', 'e', '1']
SIMPLE_STRING = 'foobar'
IMAGE_URL = 'https://hilltop-demo.s3-us-west-2.amazonaws.com/images/1557026914963.jpg'
BUCKET = 'larry-testing'
KEY = 'test-objects/s3/1557026914963.jpg'
URI = 's3://{}/{}'.format(BUCKET, KEY)
PATH_PREFIX = 's3/'
URI_PREFIX = 's3://{}/{}'.format(BUCKET, PATH_PREFIX)


class S3Tests(unittest.TestCase):

    def test_obj(self):
        o = lry.s3.obj(BUCKET, KEY)
        o.load()
        o = lry.s3.obj(URI)
        o.load()
        o = lry.s3.obj(bucket=BUCKET, key=KEY)
        o.load()
        o = lry.s3.obj(uri=URI)
        o.load()
        o = lry.s3.obj(BUCKET, KEY[:-1])
        with self.assertRaises(ClientError) as context:
            o.load()
        self.assertEqual('Not Found', context.exception.response['Error']['Message'])

    def test_delete(self):
        uri = lry.s3.fetch(IMAGE_URL, BUCKET, PATH_PREFIX + 'delete_test.jpg')
        o = lry.s3.obj(uri)
        o.load()
        lry.s3.delete(uri)
        with self.assertRaises(ClientError) as context:
            o.load()
        self.assertEqual('Not Found', context.exception.response['Error']['Message'])
        lry.s3.fetch(IMAGE_URL, BUCKET, PATH_PREFIX + 'delete_test.jpg')
        lry.s3.delete(uri=uri)
        lry.s3.fetch(IMAGE_URL, BUCKET, PATH_PREFIX + 'delete_test.jpg')
        lry.s3.delete(o.bucket_name, o.key)
        lry.s3.fetch(IMAGE_URL, BUCKET, PATH_PREFIX + 'delete_test.jpg')
        lry.s3.delete(bucket=o.bucket_name, key=o.key)
        with self.assertRaises(ClientError) as context:
            o.load()
        self.assertEqual('Not Found', context.exception.response['Error']['Message'])

    def test_delete_multiple(self):
        uris = [lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete1.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete2.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete3.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete4.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete5.txt')]
        print(uris)
        lry.s3.delete(uris)
        o = lry.s3.obj(uris[3])
        with self.assertRaises(ClientError) as context:
            o.load()
        self.assertEqual('Not Found', context.exception.response['Error']['Message'])
        uris = [lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete1.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete2.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete3.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete4.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete5.txt')]
        lry.s3.delete(uri=uris)
        uris = [lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete1.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete2.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete3.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete4.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete5.txt')]
        bucket = lry.s3.decompose_uri(uris[0])[0]
        keys = [lry.s3.decompose_uri(uri)[1] for uri in uris]
        lry.s3.delete(bucket, keys)
        uris = [lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete1.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete2.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete3.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete4.txt'),
                lry.s3.write_object(SIMPLE_STRING, BUCKET, PATH_PREFIX + 'delete5.txt')]
        lry.s3.delete(bucket=bucket, key=keys)
        o = lry.s3.obj(uris[4])
        with self.assertRaises(ClientError) as context:
            o.load()
        self.assertEqual('Not Found', context.exception.response['Error']['Message'])

    def test_get_size(self):
        self.assertGreater(lry.s3.get_size(URI), 10000)
        self.assertGreater(lry.s3.get_size(uri=URI), 10000)
        self.assertGreater(lry.s3.get_size(BUCKET, KEY), 10000)
        self.assertGreater(lry.s3.get_size(bucket=BUCKET, key=KEY), 10000)

    def test_readwrite_dict(self):
        key = PATH_PREFIX + 'dict.json'
        uri = lry.s3.compose_uri(BUCKET, key)
        dict_uri = lry.s3.write_object(SIMPLE_DICT, BUCKET, key)
        self.assertEqual(json_dumps(lry.s3.read_dict(BUCKET, key)), json_dumps(SIMPLE_DICT))
        lry.s3.delete(dict_uri)
        dict_uri = lry.s3.write_object(SIMPLE_DICT, bucket=BUCKET, key=key)
        self.assertEqual(json_dumps(lry.s3.read_dict(bucket=BUCKET, key=key)), json_dumps(SIMPLE_DICT))
        lry.s3.delete(dict_uri)
        dict_uri = lry.s3.write_object(SIMPLE_DICT, uri)
        self.assertEqual(json_dumps(lry.s3.read_dict(uri)), json_dumps(SIMPLE_DICT))
        lry.s3.delete(dict_uri)
        dict_uri = lry.s3.write_object(SIMPLE_DICT, uri=uri)
        self.assertEqual(json_dumps(lry.s3.read_dict(uri=uri)), json_dumps(SIMPLE_DICT))
        lry.s3.delete(dict_uri)

    def test_readwrite_list_of_dict(self):
        def list_dump(l):
            return [json_dumps(i) for i in l]
        key = PATH_PREFIX + 'dictlist.jsonl'
        uri = lry.s3.compose_uri(BUCKET, key)
        dictlist_uri = lry.s3.write_object(SIMPLE_LIST_OF_DICTS, BUCKET, key)
        self.assertEqual(list_dump(lry.s3.read_list_of_dict(BUCKET, key)), list_dump(SIMPLE_LIST_OF_DICTS))
        lry.s3.delete(dictlist_uri)
        dictlist_uri = lry.s3.write_object(SIMPLE_LIST_OF_DICTS, bucket=BUCKET, key=key)
        self.assertEqual(list_dump(lry.s3.read_list_of_dict(bucket=BUCKET, key=key)), list_dump(SIMPLE_LIST_OF_DICTS))
        lry.s3.delete(dictlist_uri)
        dictlist_uri = lry.s3.write_object(SIMPLE_LIST_OF_DICTS, uri)
        self.assertEqual(list_dump(lry.s3.read_list_of_dict(uri)), list_dump(SIMPLE_LIST_OF_DICTS))
        lry.s3.delete(dictlist_uri)
        dictlist_uri = lry.s3.write_object(SIMPLE_LIST_OF_DICTS, uri=uri)
        self.assertEqual(list_dump(lry.s3.read_list_of_dict(uri=uri)), list_dump(SIMPLE_LIST_OF_DICTS))
        lry.s3.delete(dictlist_uri)

    def test_readwrite_list(self):
        key = PATH_PREFIX + 'list.txt'
        uri = lry.s3.compose_uri(BUCKET, key)
        list_uri = lry.s3.write_object(SIMPLE_LIST, BUCKET, key)
        self.assertEqual(lry.s3.read_list_of_str(BUCKET, key), SIMPLE_LIST)
        lry.s3.delete(list_uri)
        list_uri = lry.s3.write_object(SIMPLE_LIST, bucket=BUCKET, key=key)
        self.assertEqual(lry.s3.read_list_of_str(bucket=BUCKET, key=key), SIMPLE_LIST)
        lry.s3.delete(list_uri)
        list_uri = lry.s3.write_object(SIMPLE_LIST, uri)
        self.assertEqual(lry.s3.read_list_of_str(uri), SIMPLE_LIST)
        lry.s3.delete(list_uri)
        list_uri = lry.s3.write_object(SIMPLE_LIST, uri=uri)
        self.assertEqual(lry.s3.read_list_of_str(uri=uri), SIMPLE_LIST)
        lry.s3.delete(list_uri)

    def test_readwrite_string(self):
        key = PATH_PREFIX + 'string.txt'
        uri = lry.s3.compose_uri(BUCKET, key)
        string_uri = lry.s3.write_object(SIMPLE_STRING, BUCKET, key)
        self.assertEqual(lry.s3.read_str(BUCKET, key), SIMPLE_STRING)
        lry.s3.delete(string_uri)
        string_uri = lry.s3.write_object(SIMPLE_STRING, bucket=BUCKET, key=key)
        self.assertEqual(lry.s3.read_str(bucket=BUCKET, key=key), SIMPLE_STRING)
        lry.s3.delete(string_uri)
        string_uri = lry.s3.write_object(SIMPLE_STRING, uri)
        self.assertEqual(lry.s3.read_str(uri), SIMPLE_STRING)
        lry.s3.delete(string_uri)
        string_uri = lry.s3.write_object(SIMPLE_STRING, uri=uri)
        self.assertEqual(lry.s3.read_str(uri=uri), SIMPLE_STRING)
        lry.s3.delete(string_uri)

    def test_rename(self):
        key1 = PATH_PREFIX + 'string1.txt'
        key2 = PATH_PREFIX + 'string2.txt'
        uri1 = lry.s3.compose_uri(BUCKET, key1)
        uri2 = lry.s3.compose_uri(BUCKET, key2)
        lry.s3.write_object(SIMPLE_STRING, BUCKET, key1)
        lry.s3.rename(old_bucket=BUCKET, old_key=key1, new_bucket=BUCKET, new_key=key2)
        self.assertEqual(lry.s3.read_str(bucket=BUCKET, key=key2), SIMPLE_STRING)
        lry.s3.delete(BUCKET, key2)
        lry.s3.write_object(SIMPLE_STRING, uri1)
        lry.s3.rename(old_uri=uri1, new_uri=uri2)
        self.assertEqual(lry.s3.read_str(uri2), SIMPLE_STRING)
        lry.s3.delete(uri2)

    def test_copy(self):
        key1 = PATH_PREFIX + 'string1.txt'
        key2 = PATH_PREFIX + 'string2.txt'
        uri1 = lry.s3.compose_uri(BUCKET, key1)
        uri2 = lry.s3.compose_uri(BUCKET, key2)
        lry.s3.write_object(SIMPLE_STRING, BUCKET, key1)
        lry.s3.copy(src_bucket=BUCKET, src_key=key1, new_bucket=BUCKET, new_key=key2)
        self.assertEqual(lry.s3.read_str(bucket=BUCKET, key=key1), lry.s3.read_str(bucket=BUCKET, key=key2))
        lry.s3.delete(BUCKET, key2)
        lry.s3.write_object(SIMPLE_STRING, uri1)
        lry.s3.copy(src_uri=uri1, new_uri=uri2)
        self.assertEqual(lry.s3.read_str(uri1), lry.s3.read_str(uri2))
        lry.s3.delete(uri2)
        lry.s3.delete(uri1)

    def test_exists(self):
        self.assertTrue(lry.s3.exists(BUCKET, KEY))
        self.assertFalse(lry.s3.exists(BUCKET, KEY[:-1]))
        self.assertTrue(lry.s3.exists(bucket=BUCKET, key=KEY))
        self.assertFalse(lry.s3.exists(bucket=BUCKET, key=KEY[:-1]))
        self.assertTrue(lry.s3.exists(URI))
        self.assertFalse(lry.s3.exists(URI[:-1]))
        self.assertTrue(lry.s3.exists(uri=URI))
        self.assertFalse(lry.s3.exists(uri=URI[:-1]))

    def test_fetch(self):
        key = PATH_PREFIX + 'fetched.jpg'
        uri = lry.s3.compose_uri(BUCKET, key)
        image_uri = lry.s3.fetch(IMAGE_URL, BUCKET, key)
        self.assertTrue(lry.s3.exists(BUCKET, key))
        lry.s3.delete(image_uri)
        image_uri = lry.s3.fetch(IMAGE_URL, bucket=BUCKET, key=key)
        self.assertTrue(lry.s3.exists(BUCKET, key))
        lry.s3.delete(image_uri)
        image_uri = lry.s3.fetch(IMAGE_URL, uri)
        self.assertTrue(lry.s3.exists(uri))
        lry.s3.delete(image_uri)
        image_uri = lry.s3.fetch(IMAGE_URL, uri=uri)
        self.assertTrue(lry.s3.exists(uri))
        lry.s3.delete(image_uri)


if __name__ == '__main__':
    unittest.main()



import setuptools
import larry as lry

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="larry",
    version=lry.__version__,
    author="Dave Schultz",
    author_email="djschult@gmail.com",
    description="Library of helper modules for common data tasks using AWS resources such as S3, SQS, MTurk and others",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dschultz0/larry",
    packages=setuptools.find_packages(),
    keywords="larry data aws boto3 mturk s3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from larry import utils
from larry import types
import boto3
import inspect
import zipfile
from io import BytesIO
import json
import builtins
import dis
from collections import Mapping
import base64
from botocore.exceptions import ClientError

client = None
# A local instance of the boto3 session to use
__session = boto3.session.Session()

TYPE_REQUEST_RESPONSE = 'RequestResponse'
TYPE_DRY_RUN = 'DryRun'
TYPE_EVENT = 'Event'


def set_session(aws_access_key_id=None,
                aws_secret_access_key=None,
                aws__session_token=None,
                region_name=None,
                profile_name=None,
                boto_session=None):
    """
    Sets the boto3 session for this module to use a specified configuration state.
    :param aws_access_key_id: AWS access key ID
    :param aws_secret_access_key: AWS secret access key
    :param aws__session_token: AWS temporary session token
    :param region_name: Default region when creating new connections
    :param profile_name: The name of a profile to use
    :param boto_session: An existing session to use
    :return: None
    """
    global __session, client
    __session = boto_session if boto_session is not None else boto3.session.Session(**utils.copy_non_null_keys(locals()))
    client = __session.client('lambda')


def get(name):
    return client.get_function(FunctionName=name)


def get_if_exists(name):
    try:
        return get(name)
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceNotFoundException':
            return None
        else:
            raise e


def create_from_function(name, function, role, runtime='python3.8', timeout=None, memory_size=None, publish=True,
                         description=None):
    package, handler = package_function(function)
    return create(name, package, handler, role, runtime, timeout, memory_size, publish, description)


def package_function(function):
    obj = BytesIO()
    zf = zipfile.ZipFile(obj, "a")
    zf.writestr('handler.py', generate_code_from_function(function), zipfile.ZIP_DEFLATED)
    for zfile in zf.filelist:
        zfile.create_system = 0
        zfile.external_attr = 0
    zf.close()
    obj.seek(0)
    return obj.read(), 'handler.' + function.__name__


def create_or_update(name, package=None, handler=None, role=None, runtime='python3.8', timeout=None, memory_size=None,
                     publish=True, description=None):
    # TODO: Add create_role=True parameter that will generate a service role with the same name
    # TODO: Ideally also inspect the code to see what boto clients it creates. In the generate step with imports?
    existing = get_if_exists(name)
    if existing:
        if package is not None:
            update(name, package, publish=publish)
        # TODO: Update the other config values if they've changed
        return existing['Configuration']['FunctionArn']
    else:
        return create(name, package, handler, role, runtime=runtime, timeout=timeout, memory_size=memory_size,
                      publish=publish, description=description)


def create(name, package, handler, role, runtime='python3.8', timeout=None, memory_size=None, publish=True,
           description=None):
    params = {
        'FunctionName': name,
        'Runtime': runtime,
        'Role': role,
        'Handler': handler,
        'Code': {'ZipFile': package},
        'Publish': publish
    }
    if description:
        params['Description'] = description
    if timeout:
        params['Timeout'] = timeout
    if memory_size:
        params['MemorySize'] = memory_size
    resp = client.create_function(**params)
    return resp['FunctionArn']


def update(name, package, publish=True, dry_run=False):
    params = utils.map_parameters(locals(), {
        'publish': 'Publish',
        'dry_run': 'DryRun',
        'name': 'FunctionName'
    })
    params['ZipFile'] = package
    resp = client.update_function_code(**params)
    return resp['FunctionArn']


def delete(name):
    client.delete_function(FunctionName=name)


def get_as_function(name, o_type=types.TYPE_DICT):
    def func(**kwargs):
        return invoke_as(name, o_type, payload=kwargs, invoke_type=TYPE_REQUEST_RESPONSE, logs=False)
    return func


def invoke(name, payload=None, invoke_type=TYPE_REQUEST_RESPONSE, logs=False, context=None):
    params = utils.map_parameters(locals(), {
        'context': 'ClientContext ',
        'invoke_type': 'InvocationType',
        'name': 'FunctionName'
    })
    if payload:
        if isinstance(payload, (Mapping, list)):
            params['Payload'] = json.dumps(payload, cls=utils.JSONEncoder)
        else:
            params['Payload'] = payload
    if logs:
        params['LogType'] = 'Tail'
        resp = client.invoke(**params)
        return resp['Payload'], base64.b64decode(resp['LogResult']).decode('utf-8').split('\n')
    else:
        resp = client.invoke(**params)
        return resp['Payload']


def invoke_as(name, o_type, payload=None, invoke_type=TYPE_REQUEST_RESPONSE, logs=False, context=None):
    result = invoke(name, payload=payload, invoke_type=invoke_type, logs=logs, context=context)
    log = None
    if logs:
        payload, log = result
    else:
        payload = result
    result = None
    if o_type == types.TYPE_STRING:
        result = payload.read().decode('utf-8')
    elif o_type == types.TYPE_DICT:
        result = json.loads(payload.read())
    else:
        raise Exception('Unhandled type')
    if logs:
        return result, log
    else:
        return result


def invoke_as_string(name, payload=None, invoke_type=TYPE_REQUEST_RESPONSE, logs=False, context=None):
    return invoke_as(name, types.TYPE_STRING, payload=payload, invoke_type=invoke_type, logs=logs, context=context)


def invoke_as_json(name, payload=None, invoke_type=TYPE_REQUEST_RESPONSE, logs=False, context=None):
    return invoke_as(name, types.TYPE_DICT, payload=payload, invoke_type=invoke_type, logs=logs, context=context)


def _get_function_calls(func, built_ins=False):
    # the used instructions
    ins = list(dis.get_instructions(func))[::-1]
    # dict for function names (so they are unique)
    names = {}

    # go through call stack
    for i, inst in list(enumerate(ins))[::-1]:
        # find last CALL_FUNCTION
        if inst.opname[:13] == "CALL_FUNCTION":

            # function takes ins[i].arg number of arguments
            ep = i + inst.arg + (2 if inst.opname[13:16] == "_KW" else 1)

            # parse argument list (Python2)
            if inst.arg == 257:
                k = i+1
                while k < len(ins) and ins[k].opname != "BUILD_LIST":
                    k += 1

                ep = k-1

            # LOAD that loaded this function
            entry = ins[ep]

            # ignore list comprehensions / ...
            name = str(entry.argval)
            if "." not in name and entry.opname == "LOAD_GLOBAL" and (built_ins or not hasattr(builtins, name)):
                # save name of this function
                names[name] = True

            # reduce this CALL_FUNCTION and all its paramters to one entry
            ins = ins[:i] + [entry] + ins[ep + 1:]
    return sorted(list(names.keys()))


def generate_code_from_function(handler, imports=('json', 'boto3')):
    code = ''

    # build the list of imports
    for val in imports:
        if val.startswith('import') or val.startswith('from'):
            code += val + '\n'
        else:
            code += 'import ' + val + '\n'

    code += '\n' + inspect.getsource(handler)

    frame = inspect.currentframe()
    try:
        ns = globals()
        cf = frame
        for fn in _get_function_calls(handler, False):
            while fn not in ns:
                cf = cf.f_back
                if cf:
                    ns = cf.f_globals
                else:
                    raise Exception('Unable to location the function "{}"'.format(fn))
            code += '\n' + inspect.getsource(ns[fn])
    finally:
        del frame
    return code

